@@ -1,5 +1,6 @@
 import json
 
+
 def main(hub, args=None):
     hub.set_quiet()
     current = hub.current
@@ -9,9 +10,16 @@ def main(hub, args=None):
 
     current = hub.current
 
-    if path[0] != "/" and not current.index:
+    if path.startswith(('http://', 'https://')):
+        url = path
+    elif path[0] != "/" and not current.index:
         hub.fatal("cannot use relative path without an active index")
-    url = current.index_url.addpath(path)
+    elif current.index:
+        url = current.index_url.addpath(path)
+    elif current.root_url:
+        url = current.root_url.addpath(path)
+    else:
+        hub.fatal("no server currently selected")
     r = hub.http_api("get", url, quiet=True, check_version=False)
     if hub.args.verbose:
         hub.line("GET REQUEST sent to %s" % url)
@@ -21,6 +29,7 @@ def main(hub, args=None):
     hub.out_json(r._json)
     return
 
+
 def main_patchjson(hub, args=None):
     hub.set_quiet()
     current = hub.current
@@ -39,4 +48,3 @@ def main_patchjson(hub, args=None):
     hub.line("PATCH REQUEST sent to %s" % url.url)
     hub.out_json(r._json)
     return
-
