@@ -3,19 +3,20 @@ import os
 import sys
 import py
 import argparse
+import shlex
 import subprocess
+import textwrap
 from base64 import b64encode
 from contextlib import closing
 from devpi import hookspecs
 from devpi_common.types import lazydecorator, cached_property
 from devpi_common.url import URL
-from devpi_common.proc import check_output
 from devpi.use import PersistentCurrent
 from devpi_common.request import new_requests_session
 from devpi import __version__ as client_version
+from pluggy import HookimplMarker
 from pluggy import PluginManager
 import json
-std = py.std
 subcommand = lazydecorator()
 
 main_description = """
@@ -24,6 +25,9 @@ packaging, uploading, installation and testing activities, using a remote
 devpi-server managed index.  For more information see http://doc.devpi.net
 """
 
+hookimpl = HookimplMarker("devpiclient")
+
+
 def main(argv=None):
     if argv is None:
         argv = list(sys.argv)
@@ -31,18 +35,22 @@ def main(argv=None):
     with closing(hub):
         return method(hub, hub.args)
 
+
 def initmain(argv):
-    args = parse_args(argv)
+    pm = get_pluginmanager()
+    args = parse_args(argv, pm)
     mod = args.mainloc
     func = "main"
     if ":" in mod:
         mod, func = mod.split(":")
     mod = __import__(mod, None, None, ["__doc__"])
-    return Hub(args), getattr(mod, func)
+    return Hub(args, pm=pm), getattr(mod, func)
 
 
 def get_pluginmanager(load_entry_points=True):
-    pm = PluginManager("devpiclient", implprefix="devpiclient_")
+    pm = PluginManager("devpiclient")
+    # support old plugins, but emit deprecation warnings
+    pm._implprefix = "devpiclient_"
     pm.add_hookspecs(hookspecs)
     if load_entry_points:
         pm.load_setuptools_entrypoints("devpi_client")
@@ -54,21 +62,24 @@ notset = object()
 
 
 class Hub:
-    class Popen(std.subprocess.Popen):
-        STDOUT = std.subprocess.STDOUT
-        PIPE = std.subprocess.PIPE
+    class Popen(subprocess.Popen):
+        STDOUT = subprocess.STDOUT
+        PIPE = subprocess.PIPE
         def __init__(self, cmds, *args, **kwargs):
             cmds = [str(x) for x in cmds]
-            std.subprocess.Popen.__init__(self, cmds, *args, **kwargs)
+            subprocess.Popen.__init__(self, cmds, *args, **kwargs)
 
-    def __init__(self, args, file=None):
+    def __init__(self, args, file=None, pm=None):
         self._tw = py.io.TerminalWriter(file=file)
         self.args = args
         self.cwd = py.path.local()
         self.quiet = False
         self._last_http_stati = []
         self.http = new_requests_session(agent=("client", client_version))
-        self.pm = get_pluginmanager()
+        if pm is None:
+            self.pm = get_pluginmanager()
+        else:
+            self.pm = pm
 
     @property
     def hook(self):
@@ -98,7 +109,8 @@ class Hub:
 
     def http_api(self, method, url, kvdict=None, quiet=False,
                  auth=notset, basic_auth=notset, cert=notset,
-                 check_version=True, fatal=True, type=None, verify=None):
+                 check_version=True, fatal=True, type=None,
+                 verify=None, headers=None):
         """ send a json request and return a HTTPReply object which
         adds some extra methods to the requests's Reply object.
 
@@ -111,11 +123,14 @@ class Hub:
         If type is specified and the json result type does not match,
         bail out fatally (unless fatal = False)
         """
-        assert kvdict is None or isinstance(kvdict, dict)
+        assert kvdict is None or isinstance(kvdict, (dict, list))
         if isinstance(url, URL):
             url = url.url
         jsontype = "application/json"
-        headers = {"Accept": jsontype, "content-type": jsontype}
+        if headers is None:
+            headers = {}
+        headers["Accept"] = jsontype
+        headers["content-type"] = jsontype
         try:
             data = json.dumps(kvdict) if kvdict is not None else None
             if auth is notset:
@@ -181,8 +196,6 @@ class Hub:
         out("%s%s %s%s" %(info, r.status_code, r.reason, message))
         return reply
 
-
-
     def requires_login(self):
         if not self.current.get_auth_user():
             self.fatal("you need to be logged in (use 'devpi login USER')")
@@ -263,10 +276,10 @@ class Hub:
 
     def popen_output(self, args, cwd=None, report=True):
         if isinstance(args, str):
-            args = std.shlex.split(args)
+            args = shlex.split(args)
         assert args[0], args
         args = [str(x) for x in args]
-        if cwd == None:
+        if cwd is None:
             cwd = self.cwd
         cmd = py.path.local.sysfind(args[0])
         if not cmd:
@@ -274,14 +287,14 @@ class Hub:
         args[0] = str(cmd)
         if report:
             self.report_popen(args, cwd)
-        return check_output(args, cwd=str(cwd))
+        return subprocess.check_output(args, cwd=str(cwd)).decode('utf-8')
 
     def popen(self, args, cwd=None, dryrun=None, **popen_kwargs):
         if isinstance(args, str):
-            args = std.shlex.split(args)
+            args = shlex.split(args)
         assert args[0], args
         args = [str(x) for x in args]
-        if cwd == None:
+        if cwd is None:
             cwd = self.cwd
         self.report_popen(args, cwd)
         if dryrun is None:
@@ -302,7 +315,7 @@ class Hub:
             rel = str(args[0])
         if extraenv is not None:
             envadd = " [%s]" % ",".join(
-                ["%s=%s" % item for item in extraenv.items()])
+                ["%s=%s" % item for item in sorted(extraenv.items())])
         else:
             envadd = ""
         self.line("--> ", base + "$", rel, " ".join(args[1:]), envadd)
@@ -408,6 +421,7 @@ def set_devpi_auth_header(headers, auth):
         auth = b64encode(auth.encode("ascii")).decode("ascii")
         headers["X-Devpi-Auth"] = auth
 
+
 class MyArgumentParser(argparse.ArgumentParser):
     class ArgumentError(Exception):
         """ and error from the argparse subsystem. """
@@ -415,12 +429,6 @@ class MyArgumentParser(argparse.ArgumentParser):
         """raise errors instead of printing and raising SystemExit"""
         raise self.ArgumentError(error)
 
-    #def __init__(self, *args, **kwargs):
-    #    kwargs["formatter_class"] = MyHelpFormatter
-    #    argparse.ArgumentParser.__init__(self, *args, **kwargs)
-
-#class MyHelpFormatter(argparse.HelpFormatter):
-#    pass
 
 def try_argcomplete(parser):
     if os.environ.get('_ARGCOMPLETE'):
@@ -451,10 +459,10 @@ def print_version(hub):
                     hub.line("    %s %s" % (name, version))
 
 
-def parse_args(argv):
+def parse_args(argv, pm):
     argv = [str(x) for x in argv]
     parser = getbaseparser('devpi')
-    add_subparsers(parser)
+    add_subparsers(parser, pm)
     try_argcomplete(parser)
     try:
         args = parser.parse_args(argv[1:])
@@ -472,6 +480,7 @@ def parse_args(argv):
         parser.print_usage()
         parser.exit(2, "%s: error: %s\n" % (parser.prog, e.args[0]))
 
+
 def parse_docstring(txt):
     description = txt
     i = txt.find(".")
@@ -481,43 +490,54 @@ def parse_docstring(txt):
         doc = txt[:i+1]
     return doc, description
 
-def add_subparsers(parser):
+
+@hookimpl
+def devpiclient_subcommands():
+    subcommands = []
+    for add_arguments, args, kwargs in subcommand.discover(globals()):
+        if len(args) > 1:
+            name = args[1]
+        else:
+            name = add_arguments.__name__
+        mainloc = args[0]
+        subcommands.append((add_arguments, name, mainloc))
+    return subcommands
+
+
+def add_subparsers(parser, pm):
     subparsers = parser.add_subparsers()
     # see http://stackoverflow.com/questions/18282403/
     # for the following two lines (py3 compat)
     subparsers.required = False
     subparsers.dest = "command"
 
-    for func, args, kwargs in subcommand.discover(globals()):
-        if len(args) > 1:
-            name = args[1]
-        else:
-            name = func.__name__
-        doc, description = parse_docstring(func.__doc__)
+    subcommands = sum(pm.hook.devpiclient_subcommands(), [])
+    for (add_arguments, name, mainloc) in subcommands:
+        doc, description = parse_docstring(add_arguments.__doc__)
         subparser = subparsers.add_parser(name,
                                           description=description,
                                           help=doc)
         subparser.Action = argparse.Action
         add_generic_options(subparser)
-        func(subparser)
-        mainloc = args[0]
+        add_arguments(subparser)
         subparser.set_defaults(mainloc=mainloc)
 
 
 def getbaseparser(prog):
     parser = MyArgumentParser(prog=prog, description=main_description)
+    if sys.version_info < (3,):
+        # workaround old argparse which doesn't support optional sub commands
+        parser.add_argument("--version", action="version",
+            version="devpi-client %s" % client_version)
+    else:
+        parser.add_argument("--version", action="store_true",
+            help="show program's version number and exit")
     add_generic_options(parser, defaults=True)
     return parser
 
+
 def add_generic_options(parser, defaults=False):
     group = parser.add_argument_group("generic options")
-    if sys.version_info < (3,):
-        # workaround old argparse which doesn't support optional sub commands
-        group.add_argument("--version", action="version",
-            version="devpi-client %s" % client_version)
-    else:
-        group.add_argument("--version", action="store_true",
-            help="show program's version number and exit")
     group.add_argument("--debug", action="store_true",
         help="show debug messages including more info on server requests")
     group.add_argument("-y", action="store_true", dest="yes",
@@ -536,24 +556,6 @@ def add_generic_options(parser, defaults=False):
         default=clientdir_default,
         help="directory for storing login and other state")
 
-@subcommand("devpi.quickstart")
-def quickstart(parser):
-    """ (deprecated) start a server, create a user and login, then create a USER/dev
-    index and then connect to this index, so that subsequent devpi
-    commands can work with it.
-    """
-    parser.add_argument("--user", action="store",
-        default=os.environ.get("USER", "test"),
-        help="set initial user name to create and login")
-    parser.add_argument("--password", action="store",
-        default="",
-        help="initial password (default is empty)")
-    parser.add_argument("--index", action="store",
-        default="dev",
-        help="initial index name for the user.")
-    parser.add_argument("--dry-run", action="store_true", dest="dryrun",
-        default=False,
-        help="don't perform any actions, just show them")
 
 @subcommand("devpi.use")
 def use(parser):
@@ -607,6 +609,7 @@ def use(parser):
              "in front of devpi-server, then use a url like this: "
              "https://username:password@example.com/USER/INDEXNAME")
 
+
 @subcommand("devpi.getjson")
 def getjson(parser):
     """ show remote server and index configuration.
@@ -618,6 +621,7 @@ def getjson(parser):
         help="path or url of a resource to show information on. "
              "examples: '/', '/user', '/user/index'.")
 
+
 @subcommand("devpi.getjson:main_patchjson")
 def patchjson(parser):
     """ send a PATCH request with the specified json content to
@@ -644,33 +648,56 @@ def list_(parser):
     """
     parser.add_argument("-f", "--failures", action="store_true",
         dest="failures",
-        help="show test setup/failure logs")
+        help="show test setup/failure logs (implies -t)")
 
     parser.add_argument("--all", action="store_true",
         help="show all versions instead of just the newest")
 
+    parser.add_argument("-t", "--toxresults", action="store_true",
+        help="show toxresults for the releases")
+
     parser.add_argument("--index", default=None,
         help="index to look at (defaults to current index)")
 
+    parser.add_argument(
+        "--ignore-bases", action="store_true", default=False,
+        help="ignore releases from inherited bases. "
+             "Requires devpi-server >= 5.0.0.")
+
     parser.add_argument("spec", nargs="?",
         help="show info for a project or a specific release. "
              "Example specs: pytest or 'pytest>=2.3.5'"
              " (Quotes are needed to prevent shell redirection)")
 
+
 @subcommand("devpi.list_remove:main_remove")
 def remove(parser):
-    """ remove project info/files from current index.
+    """\
+    removes project info/files from current index.
 
-    This command allows to remove projects or releases from your
-    current index (see "devpi use").  It will ask interactively
-    for confirmation before performing the actual removals.
+    This command allows to remove projects or releases from your current index
+    (see "devpi use").
+    It will ask interactively for confirmation before performing the actual removals.
     """
     parser.add_argument("--index", default=None,
         help="index to remove from (defaults to current index)")
-    parser.add_argument("spec",
-        help="remove info/files for a project/version/release file from the "
-             "current index. "
-             "Example specs: 'pytest' or 'pytest>=2.3.5'")
+    parser.add_argument("-f", "--force", action="store_true", default=False,
+        help="remove even on non-volatile index (with devpi-server >= 6.0.0)")
+    parser.add_argument("spec_or_url",
+        help="""\
+        describes project/version/release file(s) to release from the current index.
+        If the spec starts with 'http://' or 'https://',
+        it is considered as a request to delete a single file.""")
+    parser.formatter_class = argparse.RawDescriptionHelpFormatter
+    parser.description = textwrap.dedent(remove.__doc__)
+    parser.epilog = textwrap.dedent("""\
+        examples:
+          devpi remove pytest
+
+          devpi remove pytest>=2.3.5
+
+          devpi remove https://mydevpi.org/dev/+f/1cf/3d6eaa6cbc5fa/pytest-1.0.zip""")
+
 
 @subcommand("devpi.user")
 def user(parser):
@@ -695,7 +722,7 @@ def user(parser):
         help="user name")
     parser.add_argument("keyvalues", nargs="*", type=str,
         help="key=value configuration item.  Possible keys: "
-             "email, password.")
+             "email, password and pwhash.")
 
 
 @subcommand("devpi.user:passwd")
@@ -718,6 +745,7 @@ def login(parser):
     parser.add_argument("username", action="store", default=None,
                         help="username to use for login")
 
+
 @subcommand("devpi.login:logoff")
 def logoff(parser):
     """ log out of the current devpi-server.
@@ -756,6 +784,9 @@ def index(parser):
         help="delete an index")
     group.add_argument("-l", "--list", action="store_true",
         help="list indexes for the logged in user")
+    group.add_argument(
+        "--no-projects", action="store_true",
+        help="when showing the index data, don't include projects")
     parser.add_argument("indexname", type=str, action="store", nargs="?",
         help="index name, specified as NAME or USER/NAME.  If no index "
              "is specified use the current index")
@@ -763,6 +794,7 @@ def index(parser):
         help="key=value configuration item. Possible key=value are "
              "bases=CSV, volatile=True|False, acl_upload=CSV)")
 
+
 @subcommand("devpi.upload")
 def upload(parser):
     """ (build and) upload packages to the current devpi-server index.
@@ -786,6 +818,9 @@ def upload(parser):
     #    help="retrieve max remove version, increment and set it like --ver")
     build = parser.add_argument_group("build options")
 
+    build.add_argument(
+        "-p", "--python", default=None, metavar="PYTHON_EXE",
+        help="Specify which Python interpreter to use.")
     build.add_argument("--no-vcs", action="store_true", dest="novcs",
         help="don't VCS-export to a fresh dir, just execute setup.py scripts "
              "directly using their dirname as current dir. By default "
@@ -825,6 +860,7 @@ def upload(parser):
     #    help="answer yes on interactive questions. ")
     #
 
+
 @subcommand("devpi.test")
 def test(parser):
     """ download and test a package against tox environments.
@@ -901,7 +937,7 @@ def push(parser):
         default=None, action="store",
         help="release in format 'name==version'. of which the metadata and "
              "all release files are to be uploaded to the specified "
-             "external pypi repo." )
+             "external pypi repo.")
     parser.add_argument("target", metavar="TARGETSPEC", type=str,
         action="store",
         help="local or remote target index. local targets are of form "
@@ -930,8 +966,7 @@ def install(parser):
         help="Install from the given requirements file.")
     parser.add_argument("pkgspecs", metavar="pkg", type=str,
         action="store", default=None, nargs="*",
-        help="uri or package file for installation from current index. """
-    )
+        help="uri or package file for installation from current index. """)
 
 
 @subcommand("devpi.refresh")
@@ -963,5 +998,3 @@ def verify_reply_version(hub, reply):
     hub.fatal("devpi-client-%s got a reply with API-VERSION %s, "
               "acceptable are: %s" %(client_version, version,
                                      ",".join(acceptable_api_version)))
-
-
