@@ -13,12 +13,16 @@ else:
     vbin = "bin"
 
 devpi_endpoints = "index simpleindex pypisubmit login".split()
+devpi_data_keys = ["features"]
+
 
 def currentproperty(name):
     def propget(self):
         return self._currentdict.get(name, None)
+
     def propset(self, val):
         self._currentdict[name] = val
+
     return property(propget, propset)
 
 
@@ -33,6 +37,7 @@ class Current(object):
     _client_cert = currentproperty("client_cert")
     always_setcfg = currentproperty("always_setcfg")
     settrusted = currentproperty("settrusted")
+    features = currentproperty("features")
 
     def __init__(self):
         self._currentdict = {}
@@ -233,9 +238,19 @@ class Current(object):
         self._configure_from_server_api(r.result, url)
         # at this point we know the root url to store the following data
         if basic_auth is not None:
+            hub.warn(
+                "Use of basic authentication is deprecated, "
+                "take a look at devpi-lockdown instead. "
+                "If that doesn't work for you, "
+                "let us know by filing an issue with details of your usecase.")
             self.set_basic_auth(basic_auth[0], basic_auth[1])
         if client_cert is not None:
             if client_cert:
+                hub.warn(
+                    "Use of client side certificates is deprecated, "
+                    "take a look at devpi-lockdown instead. "
+                    "If that doesn't work for you, "
+                    "let us know by filing an issue with details of your usecase.")
                 self.set_client_cert(client_cert)
             else:
                 hub.warn("Removing stored client cert for '%s'." % url.url)
@@ -249,6 +264,9 @@ class Current(object):
             if val is not None:
                 val = rooturl.joinpath(val).url
             data[name] = val
+        for name in devpi_data_keys:
+            val = result.get(name, None)
+            data[name] = val
         self.reconfigure(data)
         status = result.get("authstatus", None)
         if status and status[0] not in ["ok", "noauth"]:
@@ -328,15 +346,16 @@ class PersistentCurrent(Current):
 
 
 def out_index_list(hub, data):
-    for user in data:
+    for user in sorted(data):
         indexes = data[user].get("indexes", {})
-        for index, ixconfig in indexes.items():
+        for index, ixconfig in sorted(indexes.items()):
             ixname = "%s/%s" % (user, index)
             hub.info("%-15s bases=%-15s volatile=%s" %(ixname,
                      ",".join(ixconfig.get("bases", [])),
                      ixconfig["volatile"]))
     return
 
+
 def active_venv():
     venv = None
     if "VIRTUAL_ENV" in os.environ:
@@ -391,6 +410,8 @@ def main(hub, args=None):
                     hub.info("%16s: %s" %(name, getattr(current, name)))
             else:
                 hub.info("current devpi index: %s (%s)" % (current.index, login_status))
+                if current.features:
+                    hub.info("supported features: %s" % ", ".join(sorted(current.features)))
         else:
             hub.info("using server: %s (%s)" % (current.rooturl, login_status))
             hub.error("no current index: type 'devpi use -l' "
@@ -437,15 +458,20 @@ def main(hub, args=None):
         show_one_conf(hub, cfg)
     hub.line("always-set-cfg: %s" % ("yes" if current.always_setcfg else "no"))
 
+
 def show_one_conf(hub, cfg):
     if not cfg.exists():
         status = "no config file exists"
     elif not cfg.indexserver:
         status = "no index server configured"
     else:
-        status = cfg.indexserver
+        url = URL(cfg.indexserver)
+        if url.password:
+            url = url.replace(password="****")
+        status = url.url
     hub.info("%-23s: %s" %(cfg.screen_name, status))
 
+
 class BaseCfg(object):
     config_name = "index_url"
     regex = re.compile(r"(index[_-]url)\s*=\s*(.*)")
@@ -505,13 +531,15 @@ class BaseCfg(object):
 
     def ensure_backup_file(self):
         if self.path.exists() and not self.backup_path.exists():
-             self.path.copy(self.backup_path)
+            self.path.copy(self.backup_path)
+
 
 class DistutilsCfg(BaseCfg):
     section_name = "[easy_install]"
     default_location = ("~/.pydistutils.cfg" if sys.platform != "win32"
                         else "~/pydistutils.cfg")
 
+
 class PipCfg(BaseCfg):
     section_name = "[global]"
 
@@ -549,7 +577,7 @@ class PipCfg(BaseCfg):
                     insection = False
             if section in line.lower() and not insection:
                 insection = True
-            if insection and re.match('index\s*=.*', line):
+            if insection and re.match(r'index\s*=.*', line):
                 line = "index = %s\n" % searchindexserver
                 found = True
             newlines.append(line)
@@ -576,7 +604,7 @@ class PipCfg(BaseCfg):
                     insection = False
             if not found and self.section_name in line.lower() and not insection:
                 insection = True
-            if not found and insection and re.match('trusted-host\s*=\s*%s' % indexserver.hostname, line):
+            if not found and insection and re.match(r'trusted-host\s*=\s*%s' % indexserver.hostname, line):
                 found = True
             newlines.append(line)
         if not found:
@@ -591,7 +619,7 @@ class PipCfg(BaseCfg):
         newlines = []
         indexserver = URL(indexserver)
         for line in self.path.readlines(cr=1):
-            if not re.match('trusted-host\s*=\s*%s' % indexserver.hostname, line):
+            if not re.match(r'trusted-host\s*=\s*%s' % indexserver.hostname, line):
                 newlines.append(line)
         self.path.write("".join(newlines))
 
@@ -603,11 +631,18 @@ class BuildoutCfg(BaseCfg):
     default_location = "~/.buildout/default.cfg"
 
 
-def parse_keyvalue_spec(keyvaluelist, keyset=None):
-    d = {}
-    for x in keyvaluelist:
-        key, val = x.split("=", 1)
-        if keyset and key not in keyset:
-            raise KeyError("invalid key: %s, allowed: %s" % (key, keyset))
-        d[key] = val
-    return d
+class KeyValues(list):
+    @property
+    def kvdict(self):
+        kvdict = {}
+        for keyvalue in self:
+            key, value = keyvalue.split("=", 1)
+            kvdict[key] = value
+        return kvdict
+
+
+def get_keyvalues(keyvaluelist):
+    for keyvalue in keyvaluelist:
+        if '=' not in keyvalue:
+            raise ValueError('No equal sign in argument')
+    return KeyValues(keyvaluelist)
